@@ -6,9 +6,6 @@ License: BSD-3
 
 '''
 
-from statsmodels.compat.python import range
-
-from collections import OrderedDict
 
 import numpy as np
 
@@ -57,7 +54,7 @@ _alias_list = [['b', 'bonf', 'bonferroni'],
                ]
 
 
-multitest_alias = OrderedDict()
+multitest_alias = {}
 for m in _alias_list:
     multitest_alias[m[0]] = m[0]
     for a in m[1:]:
@@ -74,7 +71,7 @@ def multipletests(pvals, alpha=0.05, method='hs', is_sorted=False,
         uncorrected p-values.   Must be 1-dimensional.
     alpha : float
         FWER, family-wise error rate, e.g. 0.1
-    method : string
+    method : str
         Method used for testing and adjustment of pvalues. Can be either the
         full name or initial letters. Available methods are:
 
@@ -98,13 +95,13 @@ def multipletests(pvals, alpha=0.05, method='hs', is_sorted=False,
 
     Returns
     -------
-    reject : array, boolean
+    reject : ndarray, boolean
         true for hypothesis that can be rejected for given alpha
-    pvals_corrected : array
+    pvals_corrected : ndarray
         p-values corrected for multiple tests
-    alphacSidak: float
+    alphacSidak : float
         corrected alpha for Sidak method
-    alphacBonf: float
+    alphacBonf : float
         corrected alpha for Bonferroni method
 
     Notes
@@ -279,13 +276,17 @@ def fdrcorrection(pvals, alpha=0.05, method='indep', is_sorted=False):
         set of p-values of the individual tests.
     alpha : float
         error rate
-    method : {'indep', 'negcorr')
+    method : {'indep', 'negcorr'}
+    is_sorted : bool
+        If False (default), the p_values will be sorted, but the corrected
+        pvalues are in the original order. If True, then it assumed that the
+        pvalues are already sorted in ascending order.
 
     Returns
     -------
-    rejected : array, bool
+    rejected : ndarray, bool
         True if a hypothesis is rejected, False if not
-    pvalue-corrected : array
+    pvalue-corrected : ndarray
         pvalues adjusted for multiple hypothesis testing to limit FDR
 
     Notes
@@ -367,9 +368,9 @@ def fdrcorrection_twostage(pvals, alpha=0.05, method='bky', iter=False,
 
     Returns
     -------
-    rejected : array, bool
+    rejected : ndarray, bool
         True if a hypothesis is rejected, False if not
-    pvalue-corrected : array
+    pvalue-corrected : ndarray
         pvalues adjusted for multiple hypotheses testing to limit FDR
     m0 : int
         ntest - rej, estimated number of true hypotheses
@@ -385,7 +386,7 @@ def fdrcorrection_twostage(pvals, alpha=0.05, method='bky', iter=False,
     linear step-up procedure (fdrcorrection0 with method='indep') corrected
     for the estimated fraction of true hypotheses.
     This means that the rejection decision can be obtained with
-    ``pval_corrected <= alpha``, where ``alpha`` is the origianal significance
+    ``pval_corrected <= alpha``, where ``alpha`` is the original significance
     level.
     (Note: This has changed from earlier versions (<0.5.0) of statsmodels.)
 
@@ -432,7 +433,7 @@ def fdrcorrection_twostage(pvals, alpha=0.05, method='bky', iter=False,
             break
         elif ri < ri_old:
             # prevent cycles and endless loops
-            raise RuntimeError(" oops - shouldn't be here")
+            raise RuntimeError(" oops - should not be here")
         ri_old = ri
 
     # make adjustment to pvalscorr to reflect estimated number of Non-Null cases
@@ -453,28 +454,31 @@ def fdrcorrection_twostage(pvals, alpha=0.05, method='bky', iter=False,
 
 
 def local_fdr(zscores, null_proportion=1.0, null_pdf=None, deg=7,
-              nbins=30):
+              nbins=30, alpha=0):
     """
     Calculate local FDR values for a list of Z-scores.
 
     Parameters
     ----------
-    zscores : array-like
+    zscores : array_like
         A vector of Z-scores
     null_proportion : float
         The assumed proportion of true null hypotheses
     null_pdf : function mapping reals to positive reals
         The density of null Z-scores; if None, use standard normal
-    deg : integer
+    deg : int
         The maximum exponent in the polynomial expansion of the
         density of non-null Z-scores
-    nbins : integer
+    nbins : int
         The number of bins for estimating the marginal density
         of Z-scores.
+    alpha : float
+        Use Poisson ridge regression with parameter alpha to estimate
+        the density of non-null Z-scores.
 
     Returns
     -------
-    fdr : array-like
+    fdr : array_like
         A vector of FDR values
 
     References
@@ -515,14 +519,22 @@ def local_fdr(zscores, null_proportion=1.0, null_pdf=None, deg=7,
     # The design matrix at bin centers
     dmat = np.vander(zbins, deg + 1)
 
-    # Use this to get starting values for Poisson regression
-    md = OLS(np.log(1 + zhist), dmat).fit()
+    # Rescale the design matrix
+    sd = dmat.std(0)
+    ii = sd >1e-8
+    dmat[:, ii] /= sd[ii]
+
+    start = OLS(np.log(1 + zhist), dmat).fit().params
 
     # Poisson regression
-    md = GLM(zhist, dmat, family=families.Poisson()).fit(start_params=md.params)
+    if alpha > 0:
+        md = GLM(zhist, dmat, family=families.Poisson()).fit_regularized(L1_wt=0, alpha=alpha, start_params=start)
+    else:
+        md = GLM(zhist, dmat, family=families.Poisson()).fit(start_params=start)
 
     # The design matrix for all Z-scores
     dmat_full = np.vander(zscores, deg + 1)
+    dmat_full[:, ii] /= sd[ii]
 
     # The height of the estimated marginal density of Z-scores,
     # evaluated at every observed Z-score.
@@ -552,7 +564,7 @@ class NullDistribution(object):
 
     Parameters
     ----------
-    zscores : array-like
+    zscores : array_like
         The observed Z-scores.
     null_lb : float
         Z-scores between `null_lb` and `null_ub` are all considered to be
@@ -674,11 +686,11 @@ class NullDistribution(object):
     # The fitted null density function
     def pdf(self, zscores):
         """
-        Evaluates the fitted emirical null Z-score density.
+        Evaluates the fitted empirical null Z-score density.
 
         Parameters
         ----------
-        zscores : scalar or array-like
+        zscores : scalar or array_like
             The point or points at which the density is to be
             evaluated.
 
